/*
#@HEADER
# ************************************************************************
#
#                          Moertel FE Package
#                 Copyright (2006) Sandia Corporation
#
# Under terms of Contract DE-AC04-94AL85000, there is a non-exclusive
# license for use of this work by or on behalf of the U.S. Government.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
# Questions? Contact Michael Gee (mwgee@sandia.gov)
#
# ************************************************************************
#@HEADER
*/
/* ******************************************************************** */
/* See the file COPYRIGHT for a complete copyright notice, contact      */
/* person and disclaimer.                                               */
/* ******************************************************************** */
/*!
 * \file mrtr_segment_linear1D.H
 *
 * \class MOERTEL::Segment_Linear1D
 *
 * \brief A class to define a 2-noded linear 1D Segment
 *
 * \date Last update do Doxygen: 20-March-06
 *
 */
#ifndef MOERTEL_SEGMENT_LINEAR1D_H
#define MOERTEL_SEGMENT_LINEAR1D_H

#include <ctime>
#include <iostream>
#include <map>

#include "mrtr_segment.H"
#include "mrtr_functions.H"
#include "mrtr_utils.H"

using namespace std;

/*!
\brief MOERTEL: namespace of the Moertel package

The Moertel package depends on \ref Epetra, \ref EpetraExt, \ref Teuchos,
\ref Amesos, \ref ML and \ref AztecOO:<br>
Use at least the following lines in the configure of Trilinos:<br>
\code
--enable-moertel 
--enable-epetra 
--enable-epetraext
--enable-teuchos 
--enable-ml
--enable-aztecoo --enable-aztecoo-teuchos 
--enable-amesos
\endcode

*/
namespace MOERTEL
{

/*!
\class Segment_Linear1D

\brief <b> A class to define a 2-noded linear 1D Segment</b>

This class defines a 2-noded linear 1D interface surface segment.

<b>Important:</b><br>
Upon construction, the user must assure that the nodes attached to this segment are
given in counter-clockwise order:<br>
<pre>


      <--|       o Node 1
  domain |       | 
       --        | --> outward normal n to Seg 2
                 | 
                 | Seg 2
       Seg 1     |
    o------------o Node 0
   Node 0 |  Node 1
          | 
          | outward normal n to Seg 1
          | 
          v

</pre>
The reason for this is that the outward normal to the interface is implicitly defined by
the orientation of the segments. There is no way for the MOERTEL package to check the
orientation of the segments as they are passed in by the user and most obscure results
will be produced when the orientation is uncorrect!

The \ref MOERTEL::Segment_Linear1D class supports the ostream& operator <<

\author Michael W. Gee (mwgee@sandia.gov)

*/
class  Segment_Linear1D : public MOERTEL::Segment
{
public:
  
  // @{ \name Constructors and destructors

  /*!
  \brief Standard Constructor
  
  \param Id : A unique positive Segment id. Does not need to be continous among segments
  \param nnode : Number of nodes this segment is attached to
  \param nodeId : Pointer to vector length nnode holding unique positive 
                  node ids of nodes this segment is attached to <b> in counterclock-wise
                  orientation along the interface</b>
  \param outlevel : Level of output to stdout to be generated by this class (0-10)
  */
  explicit Segment_Linear1D(int id, int nnode, int* nodeId, int out);
  
  /*!
  \brief Empty Constructor
  
  To be used for communicating Segment_Linear1D classes among processes together with 
  \ref Pack and \ref UnPack
  */
  explicit Segment_Linear1D(int out);
  
  /*!
  \brief Copy Constructor

  Makes a deep copy of a Segment_Linear1D
  
  */
  Segment_Linear1D(MOERTEL::Segment_Linear1D& old);
  
  /*!
  \brief Destructor

  */
  virtual ~Segment_Linear1D();

  /*!
  \brief Deep copy this instance of Segment_Linear1D and return pointer to the copy
  
  */
  MOERTEL::Segment* Clone();

  //@}
  // @{ \name Public members

  /*!
  \brief Pack some data from this class to an int vector of length size so
         it can be communicated using MPI
  
  */
  int* Pack(int* size);
  
  /*!
  \brief Unpack some data an int vector and store data in this class
  
  */
  bool UnPack(int* pack);
  
  /*!
  \brief Build an outward normal at segment coordinates xi
  
  */
  double* BuildNormal(double* xi);
  
  /*!
  \brief Compute and return the area of this Segment_Linear1D
  
  */
  double Area();
  
  /*!
  \brief Build the covariant basis vectors and metric tensor at a given local coord in this segment
  
  In this case of a 1D segment, there is 1 cov. basis vector and the metric tensor is a scalar
  representing its length.
  
  */
  double Metric(double* xi, double g[], double G[][3]);

  /*!
  \brief Get local coords of a node attached to this segment with \b local node Id lid 
  
  */
  bool LocalCoordinatesOfNode(int lid, double* xi);

  //@}

protected:  

  // don't want = operator
  Segment_Linear1D operator = (const Segment& old);

};

} // namespace MOERTEL

// << operator
ostream& operator << (ostream& os, const MOERTEL::Segment_Linear1D& seg);


#endif // MOERTEL_SEGMENT_LINEAR1D_H
